from datetime import date, timedelta

from wbcore import filters as wb_filters
from wbfdm.models import Instrument

from wbportfolio.filters.assets import get_latest_asset_position
from wbportfolio.models import Portfolio


class PortfolioFilterSet(wb_filters.FilterSet):
    is_tracked = wb_filters.BooleanFilter(initial=True, label="Is tracked")
    instrument = wb_filters.ModelChoiceFilter(
        label="Instrument",
        queryset=Instrument.objects.all(),
        endpoint=Instrument.get_representation_endpoint(),
        value_key=Instrument.get_representation_value_key(),
        label_key=Instrument.get_representation_label_key(),
        filter_params={"is_managed": True},
        method="filter_instrument",
    )
    modeled_after = wb_filters.ModelChoiceFilter(
        label="Modeled After",
        queryset=Portfolio.objects.all(),
        endpoint=Portfolio.get_representation_endpoint(),
        value_key=Portfolio.get_representation_value_key(),
        label_key=Portfolio.get_representation_label_key(),
        method="filter_modeled_after",
    )

    def filter_modeled_after(self, queryset, name, value):
        if value:
            modeled_after_portfolio_ids = list(
                map(
                    lambda p: p.portfolio.id, value.get_model_portfolio_relationships(date.today() - timedelta(days=7))
                )
            )
            return queryset.filter(id__in=modeled_after_portfolio_ids)
        return queryset

    def filter_instrument(self, queryset, name, value):
        if value:
            return queryset.filter(instruments=value)
        return queryset

    class Meta:
        model = Portfolio
        fields = {
            "currency": ["exact"],
            "hedged_currency": ["exact"],
            "is_manageable": ["exact"],
            "only_weighting": ["exact"],
            "is_lookthrough": ["exact"],
            "is_composition": ["exact"],
            "bank_accounts": ["exact"],
            "depends_on": ["exact"],
        }


class PortfolioTreeGraphChartFilterSet(wb_filters.FilterSet):
    date = wb_filters.DateFilter(method="fake_filter", initial=get_latest_asset_position, required=True)

    class Meta:
        model = Portfolio
        fields = {}



from contextlib import suppress
from math import ceil
from typing import Generator

from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Exists, OuterRef, Sum
from wbcompliance.models.risk_management import backend
from wbcompliance.models.risk_management.dispatch import register
from wbcore import serializers as wb_serializers
from wbfdm.models import Instrument, InstrumentPrice

from wbportfolio.models import AssetPosition


@register("Liquidity Risk", rule_group_key="portfolio")
class RuleBackend(backend.AbstractRuleBackend):
    OBJECT_FIELD_NAME: str = "instrument"

    instrument: Instrument

    def get_queryset(self):
        return self.instrument.assets.filter(
            date=self.evaluation_date, shares__isnull=False, is_invested=True
        ).exclude(shares=0)

    def is_passive_evaluation_valid(self) -> bool:
        return self.get_queryset().exists()

    @classmethod
    def get_allowed_content_type(cls) -> "ContentType":
        return ContentType.objects.get_for_model(Instrument)

    def _build_dto_args(self):
        with suppress(InstrumentPrice.DoesNotExist):
            last_price = self.instrument.valuations.get(date=self.evaluation_date)
            total_shares = self.get_queryset().aggregate(c=Sum("shares"))["c"]
            if last_price.volume_50d and total_shares:
                return float(total_shares), last_price.volume_50d
        return tuple()

    @classmethod
    def get_all_active_relationships(cls) -> models.QuerySet:
        try:
            base_qs = AssetPosition.objects.filter(shares__isnull=False, is_invested=True).exclude(shares=0)
            last_asset_position = base_qs.latest("date").date

            return Instrument.objects.annotate(
                has_assets=Exists(base_qs.filter(underlying_instrument=OuterRef("pk"), date=last_asset_position))
            ).filter(has_assets=True, children__isnull=True)
        except AssetPosition.DoesNotExist:
            return Instrument.objects.none()

    @classmethod
    def get_serializer_class(cls) -> wb_serializers.Serializer:
        class RuleBackendSerializer(wb_serializers.Serializer):
            liquidation_factor = wb_serializers.FloatField(default=3.0, label="Liquidation Factor")
            redemption_pct = wb_serializers.FloatField(default=0.80, label="Redemption Percentage", percent=True)

            @classmethod
            def get_parameter_fields(cls):
                return [
                    "liquidation_factor",
                    "redemption_pct",
                ]

        return RuleBackendSerializer

    def _process_dto(
        self, total_shares: float, volume_50d: float, **kwargs
    ) -> Generator[backend.IncidentResult, None, None]:
        days_to_liquidate = (total_shares * self.redemption_pct * self.liquidation_factor) / volume_50d
        for threshold in self.thresholds:
            numerical_range = threshold.numerical_range
            if days_to_liquidate >= numerical_range[0] and days_to_liquidate < numerical_range[1]:
                yield backend.IncidentResult(
                    breached_object=self.instrument,
                    breached_object_repr=str(self.instrument),
                    breached_value=f"{ceil(days_to_liquidate)} Days",
                    report_details={
                        "Volume 50D": volume_50d,
                        "Total Shares": total_shares,
                        "Redemption Percentage": f"{self.redemption_pct:.1%}",
                        "Liquidation Factor": self.liquidation_factor,
                    },
                    severity=threshold.severity,
                )



class OppositeSharesFieldMethodMixin:
    def filter_opposite_shares(self, queryset, name, value):
        return queryset.filter(shares=value * -1)

    def filter_opposite_approximate_shares(self, queryset, name, value):
        value = float(value)
        value_lower_bound = -0.9 * value
        value_higher_bound = -1.1 * value

        if value > 0:
            return queryset.filter(shares__lte=value_lower_bound, shares__gte=value_higher_bound)
        return queryset.filter(shares__lte=value_higher_bound, shares__gte=value_lower_bound)



import math
from datetime import datetime
from decimal import Decimal
from typing import Any, Dict

from django.db import models
from wbcore.contrib.currency.import_export.handlers import CurrencyImportHandler
from wbcore.contrib.io.exceptions import DeserializationError
from wbcore.contrib.io.imports import ImportExportHandler
from wbfdm.import_export.handlers.instrument import InstrumentImportHandler


class DividendImportHandler(ImportExportHandler):
    MODEL_APP_LABEL = "wbportfolio.DividendTransaction"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_handler = InstrumentImportHandler(self.import_source)
        self.currency_handler = CurrencyImportHandler(self.import_source)

    def _deserialize(self, data):
        data["ex_date"] = datetime.strptime(data["ex_date"], "%Y-%m-%d").date()
        data["value_date"] = datetime.strptime(data["value_date"], "%Y-%m-%d").date()
        from wbportfolio.models import Portfolio

        data["portfolio"] = Portfolio._get_or_create_portfolio(self.instrument_handler, data.get("portfolio"))
        instrument = self.instrument_handler.process_object(
            data["underlying_instrument"], only_security=False, read_only=True
        )[0]
        if not instrument:
            raise DeserializationError("Can't process this data: underlying instrument not found")
        data["underlying_instrument"] = instrument
        if "currency" not in data:
            data["currency"] = data["portfolio"].currency
        else:
            data["currency"] = self.currency_handler.process_object(data["currency"], read_only=True)[0]

        for field in self.model._meta.get_fields():
            if not (value := data.get(field.name, None)) is None and isinstance(field, models.DecimalField):
                q = 1 / (math.pow(10, 4))
                data[field.name] = Decimal(value).quantize(Decimal(str(q)))

    def _get_instance(self, data, history=None, **kwargs):
        self.import_source.log += "\nGet DividendTransaction Instance."
        self.import_source.log += f"\nParameter: Portfolio={data['portfolio']} Underlying={data['underlying_instrument']} Date={data['ex_date']}"
        dividends = history if history is not None else self.model.objects

        dividends = dividends.filter(
            portfolio=data["portfolio"],
            ex_date=data["ex_date"],
            value_date=data["value_date"],
            underlying_instrument=data["underlying_instrument"],
            price_gross=data["price_gross"],
        )
        if dividends.count() == 1:
            self.import_source.log += "\nDividendTransaction Instance Found." ""
            return dividends.first()

    def _create_instance(self, data, **kwargs):
        self.import_source.log += "\nCreate DividendTransaction."
        return self.model.objects.create(**data, import_source=self.import_source)

    def _get_history(self: models.Model, history: Dict[str, Any]) -> models.QuerySet:
        from wbportfolio.models import Product

        val_date = datetime.strptime(history["value_date"], "%Y-%m-%d")
        try:
            product = Product.objects.get(**history.get("product", {}))
            dividends = self.model.objects.filter(ex_date__lte=val_date, portfolio=product.primary_portfolio)
            if underlying_instrument_data := history.get("underlying_instrument"):
                if isinstance(underlying_instrument_data, dict):
                    dividends = dividends.filter(
                        **{f"underlying_instrument__{k}": v for k, v in underlying_instrument_data.items()}
                    )
                else:
                    dividends = dividends.filter(underlying_instrument__id=underlying_instrument_data)
            return dividends
        except Product.DoesNotExist:
            return self.models.objects.none()

    def _post_processing_history(self: models.Model, history: models.QuerySet):
        self.import_source.log += "===================="
        self.import_source.log += (
            "It was a historical import and the following DividendTransaction have to be deleted:"
        )
        for dividend in history.order_by("value_date"):
            self.import_source.log += (
                f"\n{dividend.value_date:%d.%m.%Y}: {dividend.shares} {dividend.price} ==> Deleted"
            )
            dividend.delete()



from .assets import (  # ProductAllocationEquityFilter, AssetPositionAPIFilter, AssetPositionDateFilter, ContributionChartFilter, AssetPositionIndexFilter, AssetPositionInstrumentFilter, EquityCashPositionFilter,  AssetPositionEquityFilter
    AssetPositionEquityGroupFilter,
    AssetPositionFilter,
    AssetPositionIndexGroupFilter,
    AssetPositionInstrumentFilter,
    AssetPositionPortfolioFilter,
    AssetPositionProductGroupFilter,
    AssetPositionUnderlyingInstrumentChartFilter,
    CashPositionPortfolioFilterSet,
    CompositionModelPortfolioPandasFilter,
    ContributionChartFilter,
    CompositionContributionChartFilter,
    DistributionFilter,
)
from .custodians import CustodianFilterSet
from .performances import (
    PerformanceComparisonFilter,
    PerformancePandasFilter,
    ProductPerformanceNetNewMoneyFilter,
)
from .portfolios import PortfolioFilterSet, PortfolioTreeGraphChartFilterSet
from .positions import AggregatedAssetPositionLiquidityFilter, AssetPositionPandasFilter
from .roles import PortfolioRoleFilterSet
from .signals import *
from .transactions import *
from .products import (
    BaseProductFilterSet,
    ProductFilter,
    ProductCustomerFilter,
    ProductFeeFilter,
)
from .assets_and_net_new_money_progression import AssetsAndNetNewMoneyProgressionFilterSet
from .esg import ESGMetricAggregationPortfolioPandasFilterSet



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class PortfolioRoleDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="role_type", label="Type"),
                dp.Field(key="person", label="Person"),
                dp.Field(key="start", label="Start"),
                dp.Field(key="end", label="End"),
                dp.Field(key="weighting", label="Weighting"),
                dp.Field(key="instrument", label="Instrument"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["role_type", repeat_field(2, "person")], ["start", "end", "weighting"], [repeat_field(3, "instrument")]]
        )


class PortfolioRoleInstrumentDisplayConfig(PortfolioRoleDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="role_type", label="Type"),
                dp.Field(key="person", label="Person"),
                dp.Field(key="start", label="Start"),
                dp.Field(key="end", label="End"),
                dp.Field(key="weighting", label="Weighting"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["role_type", repeat_field(2, "person")],
                ["start", "end", "weighting"],
            ]
        )



import re
from io import BytesIO

import pandas as pd
from wbfdm.models import Instrument

from wbportfolio.models import Trade

FIELD_MAP = {
    "Trade Date": "book_date",
    "Price": "price",
    "Qty": "shares",
    "Custodian": "bank",
}


def parse(import_source):
    # for extra safety we ensure that passed file satisfy some regex.
    filename_regex = import_source.source.import_parameters.get(
        "customer_trade_filename_regex",
        ".*([A-Z]{2}[A-Z0-9]{9}[0-9]{1})_TransactionList_([0-9]{4}-[0-9]{2}-[0-9]{2})_[0-9]{2}_[0-9]{2}.*.xlsx",
    )
    if match := re.match(filename_regex, import_source.file.name):
        isin = match.group(1)
        max_date = match.group(2)
        df = pd.read_excel(BytesIO(import_source.file.read()), engine="openpyxl")
        product = Instrument.objects.get(isin=isin)
        if not df.empty:
            df = df.rename(columns=FIELD_MAP)
            df = df.drop(columns=df.columns.difference(FIELD_MAP.values()))
            df["transaction_subtype"] = df.shares.apply(
                lambda x: Trade.Type.REDEMPTION if x < 0 else Trade.Type.SUBSCRIPTION
            )
            df["underlying_instrument"] = product.id
            df["book_date"] = pd.to_datetime(df["book_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
            df["portfolio"] = product.primary_portfolio.id
            df = df.dropna()
            data = df.to_dict("records")
            return {"data": data, "history": {"underlying_instrument": product.id, "book_date": max_date}}
        return {"data": {}}



import pytest
from faker import Faker
from pandas.tseries.offsets import BDay
from psycopg.types.range import NumericRange
from wbcompliance.factories.risk_management import RuleThresholdFactory

from wbportfolio.models.transactions.claim import ClaimGroupbyChoice
from wbportfolio.risk_management.backends.accounts import (
    RuleBackend as AccountRuleBackend,
)

fake = Faker()


@pytest.mark.django_db
class TestAccountRuleModel:
    @pytest.fixture
    def account_backend(
        self,
        entry,
        weekday,
        business_days_interval,
        group_by=ClaimGroupbyChoice.ACCOUNT.name,
        extra_parameters=None,
    ):
        parameters = {"group_by": group_by, "business_days_interval": business_days_interval, "field": "SHARES"}
        if extra_parameters:
            parameters.update(extra_parameters)
        return AccountRuleBackend(
            weekday,
            entry,
            parameters,
            [RuleThresholdFactory.create(range=NumericRange(lower=None, upper=-0.2))],  # detect any -20% perf
        )

    @pytest.mark.parametrize(
        "business_days_interval",
        [7],
    )
    def test_check_rule_groupby_account(
        self,
        weekday,
        entry,
        business_days_interval,
        product,
        claim_factory,
        customer_trade_factory,
        account_factory,
        account_backend,
    ):
        # Simple test to test if a valid drop in performance outside the rule window will not be detected but a one within will.
        account = account_factory.create(owner=entry)
        other_account = account_factory.create()
        claim_factory.create(
            date=(weekday - BDay(business_days_interval + 3)).date(),
            account=account,
            trade=customer_trade_factory.create(
                underlying_instrument=product, transaction_date=(weekday - BDay(business_days_interval + 3)).date()
            ),
            status="APPROVED",
            shares=100,
        )
        claim_factory.create(
            date=(weekday - BDay(business_days_interval + 2)).date(),
            account=account,
            trade=customer_trade_factory.create(
                underlying_instrument=product, transaction_date=(weekday - BDay(business_days_interval + 2)).date()
            ),
            status="APPROVED",
            shares=-50,
        )  # this drop should not be detected

        claim_factory.create(
            date=(weekday - BDay(business_days_interval + 1)).date(),
            account=account,
            trade=customer_trade_factory.create(
                underlying_instrument=product, transaction_date=(weekday - BDay(business_days_interval + 1)).date()
            ),
            status="APPROVED",
            shares=150,
        )
        claim_factory.create(
            date=(weekday - BDay(1)).date(),
            account=account,
            trade=customer_trade_factory.create(
                underlying_instrument=product, transaction_date=(weekday - BDay(1)).date()
            ),
            status="APPROVED",
            shares=-50,
        )  # this drop should be detected

        claim_factory.create(
            date=(weekday - BDay(business_days_interval + 1)).date(),
            account=other_account,
            trade=customer_trade_factory.create(
                underlying_instrument=product, transaction_date=(weekday - BDay(business_days_interval + 1)).date()
            ),
            status="APPROVED",
            shares=150,
        )
        claim_factory.create(
            date=(weekday - BDay(1)).date(),
            account=other_account,
            trade=customer_trade_factory.create(
                underlying_instrument=product, transaction_date=(weekday - BDay(1)).date()
            ),
            status="APPROVED",
            shares=-50,
        )  # this drop is valid but an another account so won't be detected
        res = list(account_backend.check_rule())
        assert len(res) == 1



from django.contrib import admin

from wbportfolio.models import Fees
from wbportfolio.models.transactions.fees import FeeCalculation


@admin.register(Fees)
class FeesAdmin(admin.ModelAdmin):
    list_filter = ["product"]
    list_display = [
        "transaction_subtype",
        "fee_date",
        "product",
        "currency",
        "total_value",
    ]
    fieldsets = (
        (
            "Fees Information",
            {
                "fields": (
                    ("transaction_subtype", "calculated", "fee_date", "product"),
                    ("currency", "currency_fx_rate"),
                    ("total_value", "total_value_fx_portfolio"),
                    ("created", "updated"),
                )
            },
        ),
    )
    autocomplete_fields = ["currency", "product"]

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("product", "currency")

    def calculate(self, request, queryset):
        for fees in queryset:
            fees.calculate_as_task.delay(fees)

    calculate.short_description = "Recalculate selected Fees"
    actions = [calculate]
    readonly_fields = ["total_value_fx_portfolio", "created", "updated"]


@admin.register(FeeCalculation)
class FeeCalculationModelAdmin(admin.ModelAdmin):
    search_fields = ("name",)



from .order_proposals import OrderProposalModelSerializer, ReadOnlyOrderProposalModelSerializer, OrderProposalRepresentationSerializer
from .orders import OrderOrderProposalModelSerializer, ReadOnlyOrderOrderProposalModelSerializer, OrderOrderProposalListModelSerializer



# TBD: Should this stay a service or should we extend the order proposal model (fat model approach) ?
from contextlib import suppress
from typing import TYPE_CHECKING

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist

from wbportfolio.order_routing import ExecutionStatus

if TYPE_CHECKING:
    from wbportfolio.models import OrderProposal


def _should_route_as_draft() -> bool:
    """Determine whether orders should be routed as drafts."""
    return getattr(settings, "DEBUG", True) or getattr(settings, "ORDER_ROUTING_AS_DRAFT", True)


def _update_orders_with_confirmations(order_proposal, confirmations, rebalancing_comment):
    """Update all orders in the proposal based on confirmed executions."""


def execute_orders(order_proposal: "OrderProposal") -> tuple[ExecutionStatus, str]:
    """
    Executes the prepared orders of an order proposal via its custodian adapter.
    Updates execution statuses and handles routing errors gracefully.
    """
    orders = order_proposal.prepare_orders_for_execution()
    as_draft = _should_route_as_draft()
    adapter = order_proposal.custodian_adapter
    order_confirmations, rebalancing_comment = adapter.submit_rebalancing(orders, as_draft=as_draft)
    leftover_orders = order_proposal.orders.all()

    for confirmed in order_confirmations:
        with suppress(ObjectDoesNotExist):
            order = leftover_orders.get(id=confirmed.id)
            order.execution_confirmed = True
            order.execution_comment = order.comment
            order.save()
            leftover_orders = leftover_orders.exclude(id=order.id)

    # Orders without confirmation
    leftover_orders.update(execution_confirmed=False, execution_comment="No confirmation received from the custodian")
    return ExecutionStatus.IN_DRAFT if as_draft else ExecutionStatus.PENDING, rebalancing_comment


def get_execution_status(order_proposal: "OrderProposal") -> tuple[ExecutionStatus, str]:
    adapter = order_proposal.custodian_adapter
    return adapter.get_rebalance_status()


def cancel_rebalancing(order_proposal: "OrderProposal") -> bool:
    adapter = order_proposal.custodian_adapter
    return adapter.cancel_current_rebalancing()



import math
from datetime import date
from decimal import Decimal

import cvxpy as cp
import numpy as np
from django.core.exceptions import ValidationError

from wbportfolio.pms.typing import Portfolio, Trade, TradeBatch


class TradeShareOptimizer:
    def __init__(self, batch: TradeBatch, portfolio_total_value: float):
        self.batch = batch
        self.portfolio_total_value = portfolio_total_value

    def optimize(self, target_cash: float = 0.99):
        try:
            return self.optimize_trade_share(target_cash)
        except ValueError:
            return self.floor_trade_share()

    def optimize_trade_share(self, target_cash: float = 0.01):
        prices_fx_portfolio = np.array([trade.price_fx_portfolio for trade in self.batch.trades])
        target_allocs = np.array([trade.target_weight for trade in self.batch.trades])

        # Decision variable: number of shares (integers)
        shares = cp.Variable(len(prices_fx_portfolio), integer=True)

        # Calculate portfolio values
        portfolio_values = cp.multiply(shares, prices_fx_portfolio)

        # Target values based on allocations
        target_values = self.portfolio_total_value * target_allocs

        # Objective: minimize absolute deviation from target values
        objective = cp.Minimize(cp.sum(cp.abs(portfolio_values - target_values)))

        # Constraints
        constraints = [
            shares >= 0,  # No short selling
            cp.sum(portfolio_values) <= self.portfolio_total_value,  # Don't exceed budget
            cp.sum(portfolio_values) >= (1.0 - target_cash) * self.portfolio_total_value,  # Use at least 99% of budget
        ]

        # Solve
        problem = cp.Problem(objective, constraints)
        problem.solve(solver=cp.CBC)

        if problem != "optimal":
            raise ValueError(f"Optimization failed: {problem.status}")

        shares_result = shares.value.astype(int)
        return TradeBatch(
            [
                trade.normalize_target(target_shares=shares_result[index])
                for index, trade in enumerate(self.batch.trades)
            ]
        )

    def floor_trade_share(self):
        return TradeBatch(
            [trade.normalize_target(target_shares=math.floor(trade.target_shares)) for trade in self.batch.trades]
        )


class TradingService:
    """
    This class represents the trading service. It can be instantiated either with the target portfolio and the effective portfolio or given a direct list of trade
    In any case, it will compute all three states
    """

    def __init__(
        self,
        trade_date: date,
        effective_portfolio: Portfolio | None = None,
        target_portfolio: Portfolio | None = None,
        total_target_weight: Decimal = Decimal("1.0"),
    ):
        self.trade_date = trade_date
        if target_portfolio is None:
            target_portfolio = Portfolio(positions=())
        if effective_portfolio is None:
            effective_portfolio = Portfolio(positions=())
        # If effective portfolio and trades batch is provided, we ensure the trade batch contains at least one trade for every position
        self.trades_batch = self.build_trade_batch(effective_portfolio, target_portfolio).normalize(
            total_target_weight
        )
        self._effective_portfolio = effective_portfolio

    @property
    def errors(self) -> list[str]:
        """
        Returned the list of errors stored during the validation process. Can only be called after is_valid
        """
        if not hasattr(self, "_errors"):
            msg = "You must call `.is_valid()` before accessing `.errors`."
            raise AssertionError(msg)
        return self._errors

    @property
    def validated_trades(self) -> list[Trade]:
        """
        Returned the list of validated trade stored during the validation process. Can only be called after is_valid
        """
        if not hasattr(self, "_validated_trades"):
            msg = "You must call `.is_valid()` before accessing `.validated_trades`."
            raise AssertionError(msg)
        return self._validated_trades

    def run_validation(self, validated_trades: list[Trade]):
        """
        Test the given value against all the validators on the field,
        and either raise a `ValidationError` or simply return.
        """
        if self._effective_portfolio:
            for trade in validated_trades:
                if (
                    trade.previous_weight
                    and trade.underlying_instrument not in self._effective_portfolio.positions_map
                ):
                    raise ValidationError("All effective position needs to be matched with a validated trade")

    def build_trade_batch(
        self,
        effective_portfolio: Portfolio,
        target_portfolio: Portfolio,
    ) -> TradeBatch:
        """
        Given combination of effective portfolio and either a trades batch or a target portfolio, ensure all theres variables are set

        Args:
            effective_portfolio: The effective portfolio
            target_portfolio: The optional target portfolio
            trades_batch: The optional trades batch

        Returns: The normalized trades batch
        """

        instruments = effective_portfolio.positions_map.copy()
        instruments.update(target_portfolio.positions_map)

        trades: list[Trade] = []
        for instrument_id, pos in instruments.items():
            previous_weight = target_weight = 0
            effective_shares = target_shares = 0
            daily_return = 0
            is_cash = False
            if effective_pos := effective_portfolio.positions_map.get(instrument_id, None):
                previous_weight = effective_pos.weighting
                effective_shares = effective_pos.shares
                daily_return = effective_pos.daily_return
                is_cash = effective_pos.is_cash
            if target_pos := target_portfolio.positions_map.get(instrument_id, None):
                target_weight = target_pos.weighting
                is_cash = target_pos.is_cash
                if target_pos.shares is not None:
                    target_shares = target_pos.shares
            trade = Trade(
                underlying_instrument=instrument_id,
                previous_weight=previous_weight,
                target_weight=target_weight,
                effective_shares=effective_shares,
                target_shares=target_shares,
                date=self.trade_date,
                instrument_type=pos.instrument_type,
                currency=pos.currency,
                price=Decimal(pos.price) if pos.price is not None else Decimal("0"),
                currency_fx_rate=Decimal(pos.currency_fx_rate),
                daily_return=Decimal(daily_return),
                portfolio_contribution=effective_portfolio.portfolio_contribution,
                is_cash=is_cash,
            )
            trades.append(trade)
        return TradeBatch(trades)

    def is_valid(self, ignore_error: bool = False) -> bool:
        """
        Validate the trade batch against a set of default rules. Populate the validated_trades and errors property.
        Ignore error by default
        Args:
            ignore_error: If true, will raise the error. False by default

        Returns: True if the trades batch is valid
        """
        if not hasattr(self, "_validated_trades"):
            self._validated_trades = []
            self._errors = []
            # Run validation for every trade. If a trade is not valid, we simply exclude it from the validated trades list
            for _, trade in self.trades_batch.trades_map.items():
                try:
                    trade.validate()
                    self._validated_trades.append(trade)
                except ValidationError as exc:
                    self._errors.append(exc.message)
            try:
                # Check the overall validity of the trade batch. If this fail, we consider all trade invalids
                self.run_validation(self._validated_trades)
            except ValidationError as exc:
                self._validated_trades = []
                self._errors.append(exc.message)

            if self._errors and not ignore_error:
                raise ValidationError(self.errors)

        return not bool(self._errors)

    def get_optimized_trade_batch(self, portfolio_total_value: float, target_cash: float):
        return TradeShareOptimizer(
            self.trades_batch, portfolio_total_value
        ).floor_trade_share()  # TODO switch to the other optimization when ready



from wbcore import filters as wb_filters

from wbportfolio.models import Custodian


class CustodianFilterSet(wb_filters.FilterSet):
    mapping = wb_filters.CharFilter(lookup_expr="icontains")

    class Meta:
        model = Custodian
        fields = {"name": ["exact", "icontains"]}



import pandas as pd


def fix_quantization_error(df: pd.Series, round_precision: int):
    df = df.round(round_precision)
    quantization_error = 1.0 - df.sum()
    largest_weight = df.idxmax()
    df.loc[largest_weight] = df.loc[largest_weight] + quantization_error
    return df



from .assets import (
    AssetPositionInstrumentModelViewSet,
    AssetPositionModelViewSet,
    AssetPositionPortfolioModelViewSet,
    CashPositionPortfolioPandasAPIView,
    CompositionModelPortfolioPandasView,
)
from .charts import (
    DistributionChartViewSet,
    DistributionTableViewSet,
    AssetPositionUnderlyingInstrumentChartViewSet,
    ContributorPortfolioChartView

)
from .custodians import CustodianModelViewSet, CustodianRepresentationViewSet
from .portfolio_relationship import InstrumentPreferedClassificationThroughProductModelViewSet
from .portfolios import (
    PortfolioModelViewSet,
    PortfolioPortfolioThroughModelViewSet,
    PortfolioRepresentationViewSet,
    PortfolioTreeGraphChartViewSet,
    TopDownPortfolioCompositionPandasAPIView
)
from .positions import (
    AggregatedAssetPositionLiquidityPandasView,
    AssetPositionPandasView,
)
from .registers import RegisterModelViewSet, RegisterRepresentationViewSet
from .roles import PortfolioRoleInstrumentModelViewSet, PortfolioRoleModelViewSet
from .rebalancing import RebalancingModelRepresentationViewSet, RebalancerRepresentationViewSet, RebalancerModelViewSet
from .transactions import *
from .orders import *
from .adjustments import AdjustmentEquityModelViewSet, AdjustmentModelViewSet
from .product_groups import ProductGroupModelViewSet, ProductGroupRepresentationViewSet
from .product_performance import (
    PerformanceComparisonPandasView,
    PerformancePandasView,
    ProductPerformanceNetNewMoneyListViewSet,
)
from .products import (
    ProductCustomerModelViewSet,
    ProductModelViewSet,
    ProductPerformanceFeesModelViewSet,
    ProductRepresentationViewSet,
    NominalProductChartView,
    AUMProductChartView,
    InstrumentPriceAUMDataChartView,
)
from .product_groups import ProductGroupRepresentationViewSet, ProductGroupModelViewSet
from .portfolio_relationship import (
    InstrumentPreferedClassificationThroughProductModelViewSet,
    InstrumentPortfolioThroughPortfolioModelViewSet,
)
from .portfolio_cash_flow import DailyPortfolioCashFlowModelViewSet
from .portfolio_swing_pricing import PortfolioSwingPricingModelViewSet
from .portfolio_cash_targets import PortfolioCashTargetModelViewSet
from .assets_and_net_new_money_progression import AssetAndNetNewMoneyProgressionChartViewSet
from .esg import ESGMetricAggregationPortfolioPandasViewSet
from .reconciliations import AccountReconciliationModelViewSet, AccountReconciliationLineModelViewSet



from typing import Optional

from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbportfolio.models import Adjustment


class AdjustmentDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="instrument", label="Equity")
                if not self.view.kwargs.get("instrument_id", None)
                else None,
                dp.Field(key="date", label="Date"),
                dp.Field(key="factor", label="Factor"),
                dp.Field(key="cumulative_factor", label="Cumulative Factor"),
                dp.Field(key="last_handler", label="Last Handler"),
            ],
            legends=[
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.BLUE_LIGHT.value,
                            label=Adjustment.Status.PENDING.label,
                            value=Adjustment.Status.PENDING.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREEN_DARK.value,
                            label=Adjustment.Status.APPLIED.label,
                            value=Adjustment.Status.APPLIED.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.RED_LIGHT.value,
                            label=Adjustment.Status.DENIED.label,
                            value=Adjustment.Status.DENIED.value,
                        ),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="status",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
                            condition=("==", Adjustment.Status.PENDING.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_DARK.value},
                            condition=("==", Adjustment.Status.APPLIED.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_LIGHT.value},
                            condition=("==", Adjustment.Status.DENIED.value),
                        ),
                    ],
                )
            ],
        )

    def get_instance_display(self) -> Display:
        if "instrument_id" in self.view.kwargs:
            fields = [
                [repeat_field(3, "status")],
                ["date", "factor", "cumulative_factor"],
            ]
        else:
            fields = [
                [repeat_field(3, "status")],
                [repeat_field(3, "instrument")],
                ["date", "factor", "cumulative_factor"],
            ]
        return create_simple_display(fields)



from .claim import (
    ClaimAccountSerializer,
    ClaimAPIModelSerializer,
    ClaimCustomerModelSerializer,
    ClaimModelSerializer,
    ClaimRepresentationSerializer,
    ClaimTradeModelSerializer,
    NegativeTermimalAccountPerProductModelSerializer,
)
from .dividends import DividendModelSerializer, DividendRepresentationSerializer
from .fees import FeesModelSerializer, FeesRepresentationSerializer
from .trades import (
    TradeModelSerializer,
    TradeRepresentationSerializer
)



from decimal import Decimal

import factory
from faker import Faker
from wbfdm.factories import InstrumentPriceFactory

from wbportfolio.models import Order

fake = Faker()


class OrderFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Order

    order_proposal = factory.SubFactory("wbportfolio.factories.OrderProposalFactory")
    currency_fx_rate = Decimal(1.0)
    fees = Decimal(0.0)
    underlying_instrument = factory.SubFactory("wbfdm.factories.InstrumentFactory")
    shares = factory.Faker("pydecimal", min_value=10, max_value=1000, right_digits=4)

    @factory.post_generation
    def create_price(self, create, extracted, **kwargs):
        if create:
            p = InstrumentPriceFactory.create(
                instrument=self.underlying_instrument, date=self.value_date, calculated=False
            )
            self.price = p.net_value
            self.save()



import datetime
from datetime import timedelta
from decimal import Decimal
from unittest.mock import patch

import pytest
from faker import Faker
from wbfdm.models import Instrument

from wbportfolio.models import Product, Trade

from ...preferences import get_product_termination_notice_period
from .utils import PortfolioTestMixin

fake = Faker()


@pytest.mark.django_db
class TestProductModel(PortfolioTestMixin):
    def test_init(self, product):
        assert product.id is not None

    def test_white_label(self, white_label_product):
        assert white_label_product.white_label_product

    def test_total_value(self, product_factory, trade_factory, instrument_price_factory):
        product = product_factory.create()
        t1 = trade_factory.create(
            shares=100,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 1) - datetime.timedelta(days=2),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        t2 = trade_factory.create(
            shares=50,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 1) - datetime.timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        t3 = trade_factory.create(  # noise
            shares=10,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        p1 = instrument_price_factory.create(
            instrument=Instrument.objects.get(id=product.id), date=t1.transaction_date
        )
        p2 = instrument_price_factory.create(
            instrument=Instrument.objects.get(id=product.id), date=t2.transaction_date
        )
        p3 = instrument_price_factory.create(
            instrument=Instrument.objects.get(id=product.id), date=t3.transaction_date
        )
        assert product.get_total_aum_usd(p1.date) == pytest.approx(Decimal(0), rel=Decimal("1e-4"))
        assert product.get_total_aum_usd(p2.date) == pytest.approx(
            Decimal(p2.net_value * (t1.shares)), rel=Decimal("1e-4")
        )
        assert product.get_total_aum_usd(p3.date) == pytest.approx(
            Decimal(p3.net_value * (t1.shares + t2.shares)), rel=Decimal("1e-4")
        )

    def test_total_shares_with_date(self, product_factory, trade_factory):
        date1 = datetime.date(2010, 1, 1)
        date2 = datetime.date(2010, 1, 2)
        product = product_factory.create()
        sub1 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date1 - datetime.timedelta(days=1),
            value_date=date1,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        sub2 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date2 - datetime.timedelta(days=1),
            value_date=date2,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        red1 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date1 - datetime.timedelta(days=1),
            value_date=date1,
            transaction_subtype=Trade.Type.REDEMPTION,
        )
        red2 = trade_factory.create(
            underlying_instrument=product,
            transaction_date=date2 - datetime.timedelta(days=1),
            value_date=date2,
            transaction_subtype=Trade.Type.REDEMPTION,
        )
        assert product.total_shares(date1 - datetime.timedelta(days=1)) == 0
        assert product.total_shares(date1) == red1.shares + sub1.shares
        assert product.total_shares(date2) == red1.shares + red2.shares + sub1.shares + sub2.shares

    def test_nominal_value_with_date(self, product_factory, trade_factory):
        product = product_factory.create()
        trade_factory.create_batch(
            5,
            shares=100,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 1) - datetime.timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        trade_factory.create_batch(
            5,
            shares=100,
            underlying_instrument=product,
            transaction_date=datetime.date(2010, 1, 2) - datetime.timedelta(days=1),
            transaction_subtype=Trade.Type.SUBSCRIPTION,
        )
        assert product.nominal_value(datetime.date(2010, 1, 1)) == (500 * product.share_price)
        assert product.nominal_value(datetime.date(2010, 1, 2)) == (1000 * product.share_price)

    def test_get_title(self, product, user):
        assert product.get_title()

    def test_compute_str(self, product, user):
        assert product.compute_str()

        # test if we save the parent instrument, the computed str is the one from the product class
        instrument = product.instrument_ptr
        instrument.save()
        product.refresh_from_db()
        assert product.computed_str == product.compute_str()

    def test_urlify_title(self, product):
        assert product.name.lower() in product.urlify_title

    def test_get_products_internal_user(self, product, white_label_product, internal_user_factory):
        assert set(Product.get_products(internal_user_factory.create().profile)) == {product, white_label_product}

    def test_get_products_normal_user(self, product_factory, user, company):
        public_product = product_factory.create()
        assert set(Product.get_products(user.profile)) == {public_product}

        white_label_product_profile = product_factory.create()
        white_label_product_profile.white_label_customers.add(user.profile)

        white_label_product_employer = product_factory.create()
        user.profile.employers.add(company)
        white_label_product_employer.white_label_customers.add(company)

        assert set(Product.get_products(user.profile)) == {
            white_label_product_employer,
            white_label_product_profile,
            public_product,
        }

    def test_subquery_is_white_label_product(self, product, white_label_product):
        product_queryset = Product.objects.all().annotate(is_white_label=Product.subquery_is_white_label_product())
        assert product_queryset.filter(is_white_label=True).count() == 1

    def test_annotate_aum(
        self, portfolio, product_factory, instrument_price_factory, trade_factory, currency_fx_rates_factory
    ):
        p1 = product_factory.create()
        d = datetime.date(2021, 1, 1)
        c1 = trade_factory.create(
            portfolio=p1.portfolio,
            underlying_instrument=p1,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            value_date=d,
            transaction_date=d,
        )
        v1 = instrument_price_factory.create(instrument=p1, date=d, outstanding_shares=c1.shares)
        fx_rate = p1.currency.fx_rates.get(date=d)
        qs = Product.annotate_last_aum(Product.objects.all(), val_date=d)
        p1.update_last_valuation_date()
        assert float(qs.first().assets_under_management_usd) == pytest.approx(
            float(v1.net_value) * float(c1.shares) / float(fx_rate.value)
        )

    def test_account_merging(self, account_factory, product):
        base_account = account_factory.create()
        merged_account = account_factory.create()
        product.default_account = merged_account
        product.save()
        base_account.merge(merged_account)
        product.refresh_from_db()
        assert product.default_account == base_account

    def test_high_water_mark(self, product, instrument_price_factory):
        initial_price = instrument_price_factory.create(instrument=product, net_value=100)
        p1 = instrument_price_factory.create(instrument=product, net_value=200)
        p2 = instrument_price_factory.create(instrument=product, net_value=100)
        p3 = instrument_price_factory.create(instrument=product, net_value=300)
        p4 = instrument_price_factory.create(instrument=product, net_value=100)
        assert product.get_high_water_mark(p1.date) == initial_price.net_value
        assert product.get_high_water_mark(p2.date) == p1.net_value
        assert product.get_high_water_mark(p3.date) == p1.net_value
        assert product.get_high_water_mark(p4.date) == p3.net_value
        product.initial_high_water_mark = 400
        product.save()
        assert product.get_high_water_mark(p4.date) == 400

    @patch("wbportfolio.models.products.send_notification")
    def test_check_and_notify_product_termination_on_date(
        self, mock_fct, weekday, product, manager_portfolio_role_factory, internal_user
    ):
        assert not product.check_and_notify_product_termination_on_date(weekday)

        product.delisted_date = weekday + datetime.timedelta(days=get_product_termination_notice_period())
        product.save()
        assert product.check_and_notify_product_termination_on_date(weekday)
        mock_fct.assert_not_called()
        assert not product.check_and_notify_product_termination_on_date(
            weekday + timedelta(days=1)
        )  # one day later is not valid anymore

        manager_portfolio_role_factory.create(person=internal_user.profile)
        product.check_and_notify_product_termination_on_date(weekday)
        mock_fct.assert_called_with(
            code="wbportfolio.product.termination_notice",
            title="Product Termination Notice",
            body=f"The product {product} will be terminated on the {product.delisted_date:%Y-%m-%d}",
            user=internal_user,
        )

    def test_delist_product_disable_report(self, product, report_factory):
        report = report_factory.create(content_object=product, is_active=True)
        assert product.delisted_date is None
        assert report.is_active

        product.delisted_date = datetime.date.today()
        product.save()

        report.refresh_from_db()
        assert report.is_active is False



import codecs
import csv
import datetime
import logging
import re

from wbportfolio.import_export.utils import convert_string_to_number
from wbportfolio.models import Fees

logger = logging.getLogger("importers.parsers.sglux.perf_fees")


def parse(import_source):
    fee_file = import_source.file.open()
    fee_file = codecs.iterdecode(fee_file, "latin1")

    # Read file into a CSV Dict Reader
    # fee_reader = csv.DictReader(fee_file, delimiter=',')
    fee_reader = csv.reader(fee_file)
    isin = re.findall("([A-Z]{2}[A-Z0-9]{9}[0-9]{1})", import_source.file.name)[0]

    # Iterate through the CSV File and parse the data into a list
    data = list()

    for fee_data in fee_reader:
        date = datetime.datetime.strptime(fee_data[0], "%m/%d/%Y")
        data.append(
            {
                "product": {"isin": isin},
                "transaction_date": date.strftime("%Y-%m-%d"),
                "calculated": False,
                "transaction_subtype": Fees.Type.PERFORMANCE,
                "total_value": round(convert_string_to_number(fee_data[1]), 6),
            }
        )
        data.append(
            {
                "product": {"isin": isin},
                "transaction_date": date.strftime("%Y-%m-%d"),
                "calculated": False,
                "transaction_subtype": Fees.Type.PERFORMANCE_CRYSTALIZED,
                "total_value": round(convert_string_to_number(fee_data[2]), 6),
            }
        )

    fee_file.close()

    return {"data": data}



from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.geography.serializers import GeographyRepresentationSerializer

from wbportfolio.models import Register


class RegisterRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="wbportfolio:register-detail")

    class Meta:
        model = Register
        fields = ("id", "computed_str", "_detail")


class RegisterModelSerializer(serializers.ModelSerializer):
    _custodian_country = GeographyRepresentationSerializer(source="custodian_country", filter_params={"level": 1})
    _custodian_city = GeographyRepresentationSerializer(source="custodian_city", filter_params={"level": 3})
    _outlet_city = GeographyRepresentationSerializer(source="outlet_city", filter_params={"level": 3})
    _outlet_country = GeographyRepresentationSerializer(source="outlet_country", filter_params={"level": 1})
    _citizenship = GeographyRepresentationSerializer(source="citizenship", filter_params={"level": 1})
    _residence = GeographyRepresentationSerializer(source="residence", filter_params={"level": 1})

    @serializers.register_only_instance_resource()
    def trades(self, instance, request, user, **kwargs):
        if instance.trades.exists():
            base_url = reverse("wbportfolio:trade-list", request=request)
            return {"trades": f"{base_url}?register={instance.id}"}
        return {}

    class Meta:
        model = Register
        fields = (
            "id",
            "register_reference",
            "register_name_1",
            "register_name_2",
            "global_register_reference",
            "external_register_reference",
            "custodian_reference",
            "custodian_name_1",
            "custodian_name_2",
            "custodian_address",
            "custodian_postcode",
            "custodian_city",
            "custodian_country",
            "sales_reference",
            "dealer_reference",
            "outlet_reference",
            "outlet_name",
            "outlet_address",
            "outlet_postcode",
            "outlet_city",
            "_outlet_city",
            "outlet_country",
            "citizenship",
            "residence",
            "investor_type",
            "status",
            "status_message",
            "opened",
            "opened_reference_1",
            "opened_reference_2",
            "updated_reference_1",
            "updated_reference_2",
            # Related Fields
            "_custodian_country",
            "_custodian_city",
            "_outlet_country",
            "_citizenship",
            "_residence",
            "_additional_resources",
        )



from datetime import date

from django.db.models import F, OuterRef, Subquery, Sum
from django.utils.functional import cached_property
from wbcore import viewsets
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import EmployerEmployeeRelationship
from wbcore.contrib.directory.serializers import NewPersonModelSerializer
from wbcore.contrib.directory.serializers import (
    PersonModelSerializer as BasePersonModelSerializer,
)
from wbcore.contrib.directory.viewsets.entries import (
    CompanyModelViewSet as OriginalCompanyModelViewSet,
)
from wbcore.contrib.directory.viewsets.entries import (
    PersonModelViewSet as OriginalPersonModelViewSet,
)
from wbcore.utils.strings import format_number

from .configs import (
    AssetAllocationDisplay,
    AssetAllocationModelEndpointConfig,
    CompanyModelDisplay,
    CompanyPreviewConfig,
    GeographicFocusDisplay,
    GeographicFocusModelEndpointConfig,
    PersonModelDisplay,
)
from .filters import CompanyFilter, PersonFilter
from .models import AssetAllocation, AssetAllocationType, GeographicFocus
from .serializers import (
    AssetAllocationModelSerializer,
    AssetAllocationTypeModelSerializer,
    AssetAllocationTypeRepresentationSerializer,
    CompanyModelListSerializer,
    CompanyModelSerializer,
    GeographicFocusModelSerializer,
    PersonModelListSerializer,
    PersonModelSerializer,
)


class CompanyModelViewSet(OriginalCompanyModelViewSet):
    LIST_DOCUMENTATION = "wbportfolio/markdown/documentation/company.md"
    display_config_class = CompanyModelDisplay
    preview_config_class = CompanyPreviewConfig
    filterset_class = CompanyFilter
    ordering_fields = OriginalCompanyModelViewSet.ordering_fields + (
        "invested_assets_under_management_usd__nulls_last",
        "potential__nulls_last",
        "asset_under_management__nulls_last",
    )

    @cached_property
    def fx_rate_date(self) -> date:
        try:
            return CurrencyFXRates.objects.latest("date").date
        except CurrencyFXRates.DoesNotExist:
            return date.today()

    def get_serializer_class(self):
        if self.get_action() in ["list", "list-metadata"]:
            return CompanyModelListSerializer
        return CompanyModelSerializer

    def get_aggregates(self, queryset, **kwargs):
        sum_potential = queryset.aggregate(sum_potential=Sum("potential")).get("sum_potential", 0.0)
        sum_assets_under_management_usd = queryset.aggregate(
            sum_assets_under_management_usd=Sum("assets_under_management_usd")
        ).get("sum_assets_under_management_usd", 0.0)
        sum_invested_assets_under_management_usd = queryset.aggregate(
            sum_invested_assets_under_management_usd=Sum("invested_assets_under_management_usd")
        ).get("sum_invested_assets_under_management_usd", 0.0)
        return {
            "potential": {"Σ": format_number(sum_potential)},
            "asset_under_management": {"Σ": format_number(sum_assets_under_management_usd)},
            "invested_assets_under_management_usd": {"Σ": format_number(sum_invested_assets_under_management_usd)},
        }

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                asset_under_management=F("portfolio_data__assets_under_management"),
                assets_under_management_currency=F("portfolio_data__assets_under_management_currency"),
                assets_under_management_currency_repr=F("portfolio_data__assets_under_management_currency__symbol"),
                fx=CurrencyFXRates.get_fx_rates_subquery(
                    self.fx_rate_date, currency="assets_under_management_currency", lookup_expr="exact"
                ),
                assets_under_management_usd=F("asset_under_management") * F("fx"),
                invested_assets_under_management_usd=F("portfolio_data__invested_assets_under_management_usd"),
                investment_discretion=F("portfolio_data__investment_discretion"),
                potential=F("portfolio_data__potential"),
                potential_currency=F("portfolio_data__potential_currency"),
            )
        )


class PersonModelViewSet(OriginalPersonModelViewSet):
    LIST_DOCUMENTATION = "wbportfolio/markdown/documentation/person.md"
    display_config_class = PersonModelDisplay
    filterset_class = PersonFilter
    ordering_fields = OriginalPersonModelViewSet.ordering_fields + (
        "invested_assets_under_management_usd__nulls_last",
        "asset_under_management__nulls_last",
        "potential",
    )
    serializer_class = PersonModelSerializer

    def get_serializer_class(self) -> BasePersonModelSerializer:
        if self.get_action() in ["list", "list-metadata"]:
            return PersonModelListSerializer
        elif "pk" not in self.kwargs:
            return NewPersonModelSerializer
        return super().get_serializer_class()

    def get_queryset(self):
        qs = super().get_queryset()
        qs = qs.annotate(
            asset_under_management=Subquery(
                EmployerEmployeeRelationship.objects.filter(primary=True, employee__id=OuterRef("id")).values(
                    "employer__portfolio_data__assets_under_management"
                )[:1],
            ),
            asset_under_management_currency_repr=Subquery(
                EmployerEmployeeRelationship.objects.filter(primary=True, employee__id=OuterRef("id")).values(
                    "employer__portfolio_data__assets_under_management_currency__key"
                )[:1],
            ),
            invested_assets_under_management_usd=Subquery(
                EmployerEmployeeRelationship.objects.filter(primary=True, employee__id=OuterRef("id")).values(
                    "employer__portfolio_data__invested_assets_under_management_usd"
                )[:1],
            ),
            potential=Subquery(
                EmployerEmployeeRelationship.objects.filter(primary=True, employee__id=OuterRef("id")).values(
                    "employer__portfolio_data__potential"
                )[:1],
            ),
            potential_currency=Subquery(
                EmployerEmployeeRelationship.objects.filter(primary=True, employee__id=OuterRef("id")).values(
                    "employer__portfolio_data__potential_currency"
                )[:1],
            ),
            assets_under_management_currency=Subquery(
                EmployerEmployeeRelationship.objects.filter(primary=True, employee__id=OuterRef("id")).values(
                    "employer__portfolio_data__assets_under_management_currency"
                )[:1],
            ),
            investment_discretion=EmployerEmployeeRelationship.objects.filter(
                primary=True, employee__id=OuterRef("id")
            ).values("employer__portfolio_data__investment_discretion")[:1],
        )
        return qs


class AssetAllocationTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = AssetAllocationType.objects.all()
    serializer_class = AssetAllocationTypeRepresentationSerializer


class AssetAllocationTypeModelViewSet(viewsets.ModelViewSet):
    queryset = AssetAllocationType.objects.all()
    serializer_class = AssetAllocationTypeModelSerializer


class AssetAllocationModelViewSet(viewsets.ModelViewSet):
    queryset = AssetAllocation.objects.all()
    serializer_class = AssetAllocationModelSerializer

    display_config_class = AssetAllocationDisplay
    endpoint_config_class = AssetAllocationModelEndpointConfig

    def get_queryset(self):
        qs = super().get_queryset()
        if company_id := self.kwargs.get("company_id", None):
            qs = qs.filter(company_id=company_id)
        return qs.select_related(
            "asset_type",
            "company",
        )


class GeographicFocusModelViewSet(viewsets.ModelViewSet):
    queryset = GeographicFocus.objects.all()
    serializer_class = GeographicFocusModelSerializer

    display_config_class = GeographicFocusDisplay
    endpoint_config_class = GeographicFocusModelEndpointConfig

    def get_queryset(self):
        if company_id := self.kwargs.get("company_id", None):
            return super().get_queryset().filter(company_id=company_id)
        return super().get_queryset()



from decimal import Decimal

import pytest
from django.forms.models import model_to_dict
from faker import Faker
from pandas.tseries.offsets import BDay
from wbfdm.import_export.handlers.instrument_price import InstrumentPriceImportHandler
from wbfdm.models import InstrumentPrice

from wbportfolio.import_export.handlers.asset_position import AssetPositionImportHandler
from wbportfolio.import_export.handlers.fees import FeesImportHandler
from wbportfolio.import_export.handlers.trade import TradeImportHandler
from wbportfolio.models import AssetPosition, Fees, Trade

fake = Faker()


@pytest.mark.django_db
class TestImportMixinModel:
    def test_import_trade(self, import_source, product, portfolio, trade_factory):
        def serialize(trade):
            data = model_to_dict(trade)
            data["transaction_date"] = trade.transaction_date.strftime("%Y-%m-%d")
            data["value_date"] = trade.value_date.strftime("%Y-%m-%d")
            data["underlying_instrument"] = {"id": product.id}
            data["portfolio"] = portfolio.id
            data["currency"] = {"key": portfolio.currency.key}
            del data["id"]
            del data["import_source"]
            return data

        trade = trade_factory.build()
        data = {"data": [serialize(trade)]}

        # Import non existing data
        TradeImportHandler(import_source).process(data)
        assert Trade.objects.count() == 1

        # Import already existing data
        # import_source.data['data'][0]['shares'] *= 2

        TradeImportHandler(import_source).process(data)
        assert Trade.objects.count() == 1

    def test_import_price(self, import_source, product, instrument_price_factory, instrument):
        def serialize(val):
            data = model_to_dict(val)
            for k, v in data.items():
                if isinstance(v, Decimal):
                    data[k] = float(v)
            data["date"] = f"{val.date:%Y-%m-%d}"
            data["instrument"] = {"instrument_type": "product", "id": product.id}
            del data["id"]
            del data["import_source"]
            del data["currency_fx_rate_to_usd"]
            return data

        val = instrument_price_factory.build(instrument=instrument)
        data = {"data": [serialize(val)]}
        handler = InstrumentPriceImportHandler(import_source)
        # Import non existing data
        handler.process(data)
        assert InstrumentPrice.objects.count() == 1

        # Import already existing data
        data["data"][0]["net_value"] *= 2
        handler.process(data)
        assert InstrumentPrice.objects.count() == 1

    def test_import_fees(self, import_source, product_factory, portfolio, cash_factory, fees_factory):
        product = product_factory.create()
        portfolio = product.primary_portfolio
        cash_factory.create(currency=portfolio.currency)

        def serialize(fees):
            data = model_to_dict(fees)
            data["fee_date"] = fees.fee_date.strftime("%Y-%m-%d")
            data["product"] = product.id
            data["currency"] = {"key": portfolio.currency.key}
            del data["calculated"]
            del data["id"]
            del data["import_source"]
            return data

        fees = fees_factory.build(calculated=False)
        data = {"data": [serialize(fees)]}
        handler = FeesImportHandler(import_source)
        # Import non existing data
        handler.process(data)
        assert Fees.objects.count() == 1

        # Import already existing data
        data["data"][0]["total_value"] *= 2
        handler.process(data)
        assert Fees.objects.count() == 1

    def _serialize_position(self, pos, instrument, underlying):
        return {
            "date": f"{pos.date:%Y-%m-%d}",
            "initial_shares": pos.initial_shares,
            "initial_price": pos.initial_price,
            "weighting": pos.weighting,
            "initial_currency_fx_rate": Decimal(1),
            "asset_valuation_date": f"{pos.asset_valuation_date:%Y-%m-%d}",
            "currency": {"key": underlying.currency.key},
            "portfolio": {"id": instrument.id, "instrument_type": instrument.security_instrument_type.key},
            "underlying_instrument": {
                "instrument_type": underlying.instrument_type.key,
                "currency": {"key": underlying.currency.key},
                "ticker": underlying.ticker,
                "name": underlying.name,
                "exchange": {"bbg_composite": "US"},
            },
        }

    @pytest.mark.parametrize("val_date", [(fake.date_object())])
    def test_import_assetposition_product(
        self,
        val_date,
        import_source,
        product_factory,
        currency,
        equity_factory,
        index_factory,
        cash_factory,
        asset_position_factory,
    ):
        val_date = (val_date - BDay(0)).date()
        product_portfolio = product_factory.create()
        instruments = [
            equity_factory.create(currency=currency, instrument_type__name="Equity"),
            cash_factory.create(currency=currency, instrument_type__name="Cash"),
            product_factory.create(currency=currency, instrument_type__name="Product"),
            index_factory.create(currency=currency, instrument_type__name="Index"),
        ]
        data = {
            "data": [
                self._serialize_position(
                    asset_position_factory.build(
                        date=val_date, underlying_instrument=instrument, weighting=Decimal("0.25")
                    ),
                    product_portfolio,
                    instrument,
                )
                for instrument in instruments
            ]
        }
        handler = AssetPositionImportHandler(import_source)

        # Import non existing data
        handler.process(data)
        for instrument in instruments:
            # we check that the position was created
            a = AssetPosition.objects.get(
                underlying_instrument=instrument, date=val_date, portfolio=product_portfolio.portfolio
            )
            assert a
            if instrument.is_cash:
                assert InstrumentPrice.objects.get(instrument=instrument, date=val_date, calculated=False, net_value=1)
            else:
                # and as this is an import handle by the assetposition handler, we expect the underlying instrument price to be created from the position initial price
                assert InstrumentPrice.objects.get(
                    instrument=instrument, date=val_date, calculated=False, net_value=a.initial_price
                )

        # Import already existing data
        data["data"][0]["initial_price"] *= 2

        handler.process(data)
        assert AssetPosition.objects.count() == 4

    def test_import_assetposition_product_group(
        self, import_source, product_group, currency, equity, asset_position_factory
    ):
        positions = asset_position_factory.build(underlying_instrument=equity, weighting=Decimal("1"))
        data = {"data": [self._serialize_position(positions, product_group, equity)]}

        # Import non existing data
        handler = AssetPositionImportHandler(import_source)

        handler.process(data)
        assert AssetPosition.objects.count() == 1

    def test_import_assetposition_index(
        self, import_source, index, portfolio, currency, equity, asset_position_factory
    ):
        positions = asset_position_factory.build(underlying_instrument=equity, weighting=Decimal("1"))
        index.portfolios.add(portfolio)
        data = {"data": [self._serialize_position(positions, index, equity)]}

        # Import non existing data
        handler = AssetPositionImportHandler(import_source)

        handler.process(data)
        assert AssetPosition.objects.count() == 1



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

CustodianMenuItem = MenuItem(
    label="Custodian",
    endpoint="wbportfolio:custodian-list",
    permission=ItemPermission(
        permissions=["wbportfolio.view_custodian"],
        method=lambda request: is_internal_user(request.user),
    ),
)



from contextlib import suppress
from datetime import date, timedelta
from decimal import Decimal
from typing import Generator

import pandas as pd
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Sum
from pandas.tseries.offsets import BDay
from wbcompliance.models.risk_management import backend
from wbcore import serializers as wb_serializers
from wbfdm.backends.dto import PriceDTO
from wbfdm.models import Instrument, InstrumentPrice, InstrumentType
from wbfdm.serializers import (
    InstrumentTypeRepresentationSerializer,
    SecurityRepresentationSerializer,
)

from wbportfolio.models import AssetPosition, InstrumentPortfolioThroughModel, Portfolio


class ActivePortfolioRelationshipMixin(backend.AbstractRuleBackend):
    OBJECT_FIELD_NAME: str = "portfolio"

    portfolio: Portfolio

    def is_passive_evaluation_valid(self) -> bool:
        return self.portfolio.imported_assets.filter(date=self.evaluation_date).exists()

    @classmethod
    def get_allowed_content_type(cls) -> "ContentType":
        return ContentType.objects.get_for_model(Portfolio)

    def _build_dto_args(self):
        return (self.portfolio._build_dto(self.evaluation_date, is_estimated=False),)

    @classmethod
    def get_all_active_relationships(cls) -> models.QuerySet:
        valid_relationships = InstrumentPortfolioThroughModel.objects.filter(
            instrument__instrument_type__key="product"
        ).values("portfolio")
        return Portfolio.objects.filter(id__in=valid_relationships, is_tracked=True)


class ActiveProductRelationshipMixin(backend.AbstractRuleBackend):
    OBJECT_FIELD_NAME: str = "product"

    @classmethod
    def get_allowed_content_type(cls) -> "ContentType":
        return ContentType.objects.get_for_model(Instrument)

    @classmethod
    def get_all_active_relationships(cls) -> models.QuerySet:
        return Instrument.active_objects.filter(instrument_type__key="product")


class StopLossMixin(backend.AbstractRuleBackend):
    class FreqChoices(models.TextChoices):
        BUSINESS_DAY = "B", "Business Day"
        WEEKLY_FRIDAY = "W-FRI", "Friday to Friday"
        BUSINESS_MONTHLY = "BME", "Business Monthly"
        BUSINESS_YEARLY = "BYE", "Business Yearly"

    class DateIntervalOption(models.TextChoices):
        ROLLING_WINDOWS = "ROLLING_WINDOWS", "Rolling Window"
        FREQUENCY = "FREQUENCY", "Frequency"

    class DynamicBenchmarkType(models.TextChoices):
        PORTFOLIO = "PORTFOLIO", "Primary Portfolio"
        PRIMARY_BENCHMARK = "PRIMARY_BENCHMARK", "Primary Benchmark"

    class FieldType(models.TextChoices):
        OUTSTANDING_SHARES = "outstanding_shares", "Outstanding shares"
        NET_VALUE = "net_value", "Net Value"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.portfolio = self.product.portfolio

    @classmethod
    def get_serializer_class(cls) -> wb_serializers.Serializer:
        class RuleBackendSerializer(wb_serializers.Serializer):
            freq = wb_serializers.ChoiceField(
                choices=cls.FreqChoices,
                default=cls.FreqChoices.WEEKLY_FRIDAY,
                label="Frequency",
                help_text="Valid only if the interval Option is Frequency. Specify the frequency use for aggregation",
            )
            date_interval_option = wb_serializers.ChoiceField(
                choices=cls.DateIntervalOption,
                default=cls.DateIntervalOption.ROLLING_WINDOWS,
                label="Interval Option",
            )
            dynamic_benchmark_type = wb_serializers.ChoiceField(
                choices=cls.DynamicBenchmarkType,
                default=None,
                allow_null=True,
                label="Benchmark Type",
                help_text="If specified, will compare the stop loss against the instrument related potential benchmark performance",
            )
            rolling_window_interval = wb_serializers.IntegerField(
                default=7,
                label="Rolling Window Interval",
                help_text="Valid only if interval option is Rolling Window. Specify the number of day for sampling",
            )
            static_benchmark = wb_serializers.PrimaryKeyRelatedField(
                queryset=Instrument.objects.all(),
                default=None,
                allow_null=True,
                label="Static Benchmark (If any)",
                help_text="If specified, will compare the stop loss against this benchmark",
            )
            asset_class = wb_serializers.PrimaryKeyRelatedField(
                queryset=InstrumentType.objects.all(),
                default=None,
                allow_null=True,
                label="Only Asset Class",
            )
            _asset_class = InstrumentTypeRepresentationSerializer(source="asset_class")

            _static_benchmark = SecurityRepresentationSerializer(
                source="static_benchmark", default=None, allow_null=True
            )
            field = wb_serializers.ChoiceField(
                choices=cls.FieldType.choices,
                default=cls.FieldType.NET_VALUE,
                label="Field",
            )
            penny_stock_max_abs_net_value = wb_serializers.FloatField(
                default=0.001, label="Penny Stock Max Absolute Net value"
            )

            @classmethod
            def get_parameter_fields(cls):
                return [
                    "freq",
                    "date_interval_option",
                    "dynamic_benchmark_type",
                    "rolling_window_interval",
                    "static_benchmark",
                    "asset_class",
                    "field",
                    "penny_stock_max_abs_net_value",
                ]

        return RuleBackendSerializer

    @property
    def benchmark(self):
        if self.dynamic_benchmark_type == self.DynamicBenchmarkType.PRIMARY_BENCHMARK.name:
            return self.product.primary_benchmark
        elif self.dynamic_benchmark_type == self.DynamicBenchmarkType.PORTFOLIO.name:
            return self.product
        elif self.static_benchmark:
            return self.static_benchmark

    def is_passive_evaluation_valid(self) -> bool:
        try:
            last_price = self.product.get_price(self.evaluation_date)
            base_condition = last_price != 0
            if (benchmark := self.benchmark) and (last_benchmark_price := benchmark.get_price(self.evaluation_date)):
                return (last_benchmark_price != 0) and base_condition
            return base_condition
        except ValueError:
            return False

    def _generate_incidents(
        self,
        tested_instrument_id: int,
        perf_instrument: float,
        perf_benchmark: float,
    ) -> Generator[backend.IncidentResult, None, None]:
        total_perf = perf_instrument if perf_benchmark is None else perf_instrument - perf_benchmark
        field_label = self.FieldType(self.field).label
        for threshold in self.thresholds:
            if threshold.is_inrange(total_perf):
                instrument = Instrument.objects.get(id=tested_instrument_id)
                report_details = {}
                with suppress(AssetPosition.DoesNotExist):
                    previous_portfolio_date = (
                        self.product.portfolio.assets.filter(date__lt=self.evaluation_date).latest("date").date
                    )
                    weighting = self.product.portfolio.assets.filter(
                        underlying_quote=instrument, date=previous_portfolio_date
                    ).aggregate(s=Sum("weighting"))["s"] or Decimal("0.0")
                    report_details["Portfolio Impact"] = f"{float(weighting) * total_perf:+,.2%}"
                report_details["Field"] = field_label
                if self.benchmark:
                    report_details[f"Relative Percentage VS {str(self.benchmark)}"] = f"{total_perf:,.3%}"
                if total_perf < 0:
                    breached_value = f'<span style="color:red">{total_perf:+,.2%}</span>'
                else:
                    breached_value = f'<span style="color:green">{total_perf:+,.2%}</span>'
                yield backend.IncidentResult(
                    breached_object=instrument,
                    breached_object_repr=str(instrument),
                    breached_value=breached_value,
                    report_details=report_details,
                    severity=threshold.severity,
                )

    def _get_start_interval(self) -> date:
        if self.date_interval_option == self.DateIntervalOption.FREQUENCY.name:
            return pd.date_range(end=self.evaluation_date, periods=2, freq=self.FreqChoices(self.freq).value)[0].date()
        else:
            return (self.evaluation_date - timedelta(days=self.rolling_window_interval - 1) - BDay(1)).date()

    def _get_performance(self, valuation_dto: PriceDTO) -> float:
        if not valuation_dto:
            return 0.0
        start_date = self._get_start_interval()
        if (
            qs := InstrumentPrice.objects.filter(
                calculated=False,
                instrument_id=valuation_dto.instrument,
                date__lte=start_date,
                date__gte=start_date - BDay(2),  # we allow 2 business day interval in case of market holiday
            ).exclude(
                net_value__lte=self.penny_stock_max_abs_net_value, net_value__gte=-self.penny_stock_max_abs_net_value
            )
        ).exists():
            last_value = float(getattr(qs.latest("date"), self.field))
            return float(getattr(valuation_dto, self.field)) / last_value - 1
        return 0.0

    def _build_dto_args(self) -> tuple[PriceDTO, PriceDTO | None]:
        if benchmark := self.benchmark:
            return self.product._build_dto(self.evaluation_date), benchmark._build_dto(self.evaluation_date)
        return self.product._build_dto(self.evaluation_date), None



from datetime import date, timedelta

from psycopg.types.range import DateRange

from wbportfolio.models import Fees, Trade


def get_transaction_gte_default(field, request, view):
    filter_date = date.today() - timedelta(days=90)
    qs = Trade.objects.none()
    if "instrument_id" in view.kwargs:
        qs = Trade.objects.filter(underlying_instrument__id=view.kwargs["instrument_id"])
    elif "portfolio_id" in view.kwargs:
        qs = Trade.objects.filter(portfolio__id=view.kwargs["portfolio_id"])
    if qs.exists():
        filter_date = qs.earliest("transaction_date").transaction_date
    return filter_date


def get_transaction_underlying_type_choices(*args):
    models = [Fees, Trade]
    choices = []
    for model in models:
        for choice in model.Type.choices:
            choices.append(choice)
    return choices


def get_transaction_lte_default(field, request, view):
    filter_date = date.today() + timedelta(days=7)
    qs = Trade.objects.none()
    if "instrument_id" in view.kwargs:
        qs = Trade.objects.filter(underlying_instrument__id=view.kwargs["instrument_id"])
    elif "portfolio_id" in view.kwargs:
        qs = Trade.objects.filter(portfolio__id=view.kwargs["portfolio_id"])
    if qs.exists():
        filter_date = qs.latest("transaction_date").transaction_date
    return filter_date


def get_transaction_default_date_range(*args, **kwargs):
    return DateRange(get_transaction_gte_default(*args, **kwargs), get_transaction_lte_default(*args, **kwargs))



# Generated by Django 5.0.6 on 2024-06-18 08:48

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0061_portfolio_bank_accounts_product_bank_account_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="dailyportfoliocashflow",
            options={
                "permissions": [("administrate_dailyportfoliocashflow", "Can administrate Daily Portfolio CashFlow")],
                "verbose_name": "Daily Portfolio CashFlow",
                "verbose_name_plural": "Daily Portfolio CashFlow",
            },
        ),
    ]



import logging
from contextlib import suppress
from datetime import date
from decimal import Decimal, InvalidOperation
from typing import TYPE_CHECKING

from django.contrib import admin
from django.db import models
from django.db.models import (
    Case,
    CharField,
    Exists,
    ExpressionWrapper,
    F,
    FloatField,
    OuterRef,
    Q,
    QuerySet,
    Subquery,
    Sum,
    Value,
    When,
    Window,
)
from django.db.models.functions import Coalesce
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.functional import cached_property
from pandas._libs.tslibs.offsets import BDay
from wbcore.contrib.currency.models import Currency, CurrencyFXRates
from wbcore.contrib.io.mixins import ImportMixin
from wbcore.signals import pre_merge
from wbfdm.models import Classification, ClassificationGroup, Instrument
from wbfdm.models.instruments.instrument_prices import InstrumentPrice
from wbfdm.signals import add_instrument_to_investable_universe

from wbportfolio.import_export.handlers.asset_position import AssetPositionImportHandler
from wbportfolio.models.portfolio_relationship import (
    InstrumentPortfolioThroughModel,
    PortfolioInstrumentPreferredClassificationThroughModel,
)
from wbportfolio.models.roles import PortfolioRole
from wbportfolio.pms.typing import Position as PositionDTO

logger = logging.getLogger("pms")

MARKETCAP_S = 2_000_000_000
MARKETCAP_M = 10_000_000_000
MARKETCAP_L = 50_000_000_000
MARKETCAP_XL = 300_000_000_000
LIQUIDITY_SMALL = 3.0
LIQUIDITY_LARGE = 5.0

MINUTE = 60
HOUR = MINUTE * 60
DAY = HOUR * 24

if TYPE_CHECKING:
    pass


class AssetPositionDefaultQueryset(QuerySet):
    def annotate_classification_for_group(
        self, classification_group: ClassificationGroup, classification_height: int = 0, **kwargs
    ) -> QuerySet:
        return classification_group.annotate_queryset(self, classification_height, "underlying_instrument", **kwargs)

    def annotate_preferred_classification_for_group(
        self, classification_group: ClassificationGroup, classification_height: int = 0
    ) -> QuerySet:
        ref_title = f"classification__{'parent__' * classification_height}name"
        ref_code = f"classification__{'parent__' * classification_height}code_aggregated"

        base_qs = PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(
            classification_group=classification_group,
            instrument=OuterRef("underlying_instrument"),
            instrument__tree_id=models.OuterRef("underlying_instrument__tree_id"),
            instrument__lft__lte=models.OuterRef("underlying_instrument__lft"),
            instrument__rght__gte=models.OuterRef("underlying_instrument__rght"),
            portfolio=models.OuterRef("portfolio"),
        )
        return self.annotate(
            classification_id=Subquery(base_qs.values(ref_code)[:1]),
            classification_title=Subquery(base_qs.values(ref_title)[:1]),
        )

    def annotate_hedged_currency_fx_rate(self, hedged_currency: Currency | None) -> QuerySet:
        return self.annotate(
            _is_hedged=Case(
                When(
                    underlying_instrument__currency__isnull=False,
                    underlying_instrument__currency=hedged_currency,
                    then=Value(True),
                ),
                default=Value(False),
                output_field=models.BooleanField(),
            ),
            hedged_currency_fx_rate=Case(
                When(_is_hedged=True, then=Value(Decimal(1.0))),
                default=F("currency_fx_rate"),
                output_field=models.BooleanField(),
            ),
        )


class DefaultAssetPositionManager(models.Manager):
    def annotate_classification_for_group(
        self, classification_group, classification_height: int = 0, **kwargs
    ) -> QuerySet:
        return self.get_queryset().annotate_classification_for_group(
            classification_group, classification_height=classification_height, **kwargs
        )

    def annotate_preferred_classification_for_group(
        self, classification_group, classification_height: int = 0
    ) -> QuerySet:
        return self.get_queryset().annotate_preferred_classification_for_group(
            classification_group, classification_height=classification_height
        )

    def annotate_hedged_currency_fx_rate(self, hedged_currency: Currency | None) -> QuerySet:
        return self.get_queryset().annotate_hedged_currency_fx_rate(hedged_currency)

    def get_queryset(self) -> AssetPositionDefaultQueryset:
        return AssetPositionDefaultQueryset(self.model).annotate(
            adjusting_factor=Coalesce(
                F("applied_adjustment__cumulative_factor") * F("applied_adjustment__factor"), Decimal(1.0)
            ),
            shares=F("initial_shares") / F("adjusting_factor"),
            price=Coalesce(F("underlying_quote_price__net_value"), F("initial_price")),
            market_capitalization=ExpressionWrapper(
                Coalesce(
                    F("underlying_quote_price__market_capitalization_consolidated"),
                    F("underlying_quote_price__market_capitalization"),
                ),
                output_field=models.DecimalField(),
            ),
            beta=ExpressionWrapper(F("underlying_quote_price__beta"), output_field=models.DecimalField()),
            correlation=ExpressionWrapper(
                F("underlying_quote_price__correlation"), output_field=models.DecimalField()
            ),
            sharpe_ratio=ExpressionWrapper(
                F("underlying_quote_price__sharpe_ratio"), output_field=models.DecimalField()
            ),
            volume=Coalesce(
                ExpressionWrapper(F("underlying_quote_price__volume"), output_field=models.DecimalField()),
                Decimal(0),
            ),
            volume_50d=Coalesce(
                ExpressionWrapper(F("underlying_quote_price__volume_50d"), output_field=models.DecimalField()),
                Decimal(0),
            ),
            volume_200d=Coalesce(
                ExpressionWrapper(F("underlying_quote_price__volume_200d"), output_field=models.DecimalField()),
                Decimal(0),
            ),
            currency_fx_rate_instrument_to_usd_rate=Case(
                When(currency_fx_rate_instrument_to_usd__value=0, then=Value(Decimal(1.0))),
                default=1 / F("currency_fx_rate_instrument_to_usd__value"),
            ),
            currency_fx_rate_portfolio_to_usd_rate=F("currency_fx_rate_portfolio_to_usd__value"),
            currency_fx_rate=Coalesce(
                F("currency_fx_rate_portfolio_to_usd_rate") * F("currency_fx_rate_instrument_to_usd_rate"),
                F("initial_currency_fx_rate"),
            ),
            currency_symbol=F("currency__symbol"),
            portfolio_currency_symbol=F("portfolio__currency__symbol"),
            price_fx_usd=F("price") * F("currency_fx_rate_instrument_to_usd_rate"),
            price_fx_portfolio=F("price") * F("currency_fx_rate"),
            total_value=F("price") * F("shares"),
            total_value_fx_usd=F("price") * F("shares") * F("currency_fx_rate_instrument_to_usd_rate"),
            total_value_fx_portfolio=F("price") * F("shares") * F("currency_fx_rate"),
            market_share=Case(
                When(market_capitalization=0, then=Value(None)),
                default=F("total_value") / F("market_capitalization"),
            ),
            liquidity=Case(
                When(volume_50d=0, then=Value(None)),
                default=ExpressionWrapper(F("shares") / F("volume_50d") / 0.33, output_field=FloatField()),
            ),
            market_capitalization_usd=F("market_capitalization") * F("currency_fx_rate_instrument_to_usd_rate"),
            volume_usd=ExpressionWrapper(
                (F("price_fx_portfolio") * F("currency_fx_rate_instrument_to_usd_rate") * F("volume_50d")),
                output_field=FloatField(),
            ),
            is_invested=Case(
                When(
                    Q(portfolio__invested_timespan__startswith__lte=F("date"))
                    & Q(portfolio__invested_timespan__endswith__gt=F("date")),
                    then=Value(True),
                ),
                default=Value(False),
                output_field=models.BooleanField(),
            ),
        )


class AnalyticalAssetPositionManager(DefaultAssetPositionManager):
    def get_queryset(self):
        qs_default = super().get_queryset()
        return qs_default.annotate(
            last_portfolio_date=Subquery(
                qs_default.filter(
                    portfolio=OuterRef("portfolio"),
                    date__lt=OuterRef("date"),
                )
                .order_by("-date")
                .values("date")[:1]
            ),
            previous_price_usd=Subquery(
                qs_default.filter(
                    date=OuterRef("last_portfolio_date"),
                    underlying_quote=OuterRef("underlying_quote"),
                    portfolio=OuterRef("portfolio"),
                )
                .order_by("-date")
                .values("price_fx_usd")[:1]
            ),
            performance=ExpressionWrapper(F("price_fx_usd") / F("previous_price_usd") - 1, output_field=FloatField()),
            contribution=ExpressionWrapper(F("performance") * F("weighting"), output_field=FloatField()),
            cumulative_contribution=Window(Sum("contribution"), order_by=F("date").asc()),
        )


class AssetPositionGroupBy(models.TextChoices):
    INDUSTRY = "classification", "Industry"
    INSTRUMENT_TYPE = "instrument_type", "Type"
    COUNTRY = "country", "Country"
    CURRENCY = "currency", "Currency"
    CASH = "is_cash", "Cash"


class AssetPosition(ImportMixin, models.Model):
    """
    The Asset Model holds all information needed to compute the value of the asset and contribution to its wbportfolio. All
    information which are regarded as Meta-Information, such as country, industry, currency allocation is held in the
    asset information, which is accessible through a FK, depending on the asset type of the asset.
    """

    import_export_handler_class = AssetPositionImportHandler
    is_estimated = models.BooleanField(
        default=False,
        verbose_name="Estimated Asset",
        help_text="True if the data is "
        "forward-estimated "
        "based on last day "
        "data. If the data is "
        "overridden by the "
        "importer or by the "
        "synchronization, "
        "this field becomes "
        "False.",
    )

    date = models.DateField(default=date.today)

    initial_price = models.DecimalField(max_digits=16, decimal_places=4, verbose_name="Initial Price")
    initial_currency_fx_rate = models.DecimalField(
        decimal_places=14,
        max_digits=28,
        verbose_name="Initial Currency FX Rate",
        help_text="The Currency Exchange Rate that is applied to the Asset to convert it into the Portfolio's currency.",
        default=Decimal(1),
    )
    initial_shares = models.DecimalField(
        decimal_places=4,
        max_digits=18,
        null=True,
        blank=True,
        verbose_name="Initial Quantity",
        help_text="The amount of Units of the Asset on the price date of the Asset.",
    )
    applied_adjustment = models.ForeignKey(
        "wbportfolio.Adjustment", on_delete=models.SET_NULL, blank=True, null=True, related_name="adjustmented_assets"
    )

    asset_valuation_date = models.DateField(
        verbose_name="Alternate Valuation Date",
        help_text="An alternate Valuation Date, if the price date of the Asset is different from the overlying Portfolio.",
    )

    exchange = models.ForeignKey(
        to="wbfdm.Exchange",
        null=True,
        blank=True,
        related_name="assets",
        on_delete=models.PROTECT,
        verbose_name="Exchange",
        help_text="The exchange where this asset is.",
    )

    ########################################################
    #                      VALUATION                       #
    ########################################################

    class PriceDenotation(models.TextChoices):
        PERCENT = "PERCENT", "%"
        CURRENCY = "CURRENCY", "Currency"

    price_denotation = models.CharField(
        max_length=16,
        choices=PriceDenotation.choices,
        default=PriceDenotation.CURRENCY,
        verbose_name="Price Denotation",
        help_text="The denotation of the price.",
    )

    # adjusting_factor = models.DecimalField(
    #     decimal_places=4,
    #     max_digits=16,
    #     default=Decimal(1),
    #     verbose_name="Share Multiplier",
    #     help_text="The Share Multiplier of an Asset.",
    # )

    weighting = models.DecimalField(
        decimal_places=8,
        max_digits=9,
        default=Decimal(0),
        verbose_name="Weight",
        help_text="The Weight of the Asset on the price date of the Asset.",
    )

    ########################################################
    #                      PORTFOLIO                       #
    ########################################################

    portfolio = models.ForeignKey(
        to="wbportfolio.Portfolio",
        related_name="assets",
        on_delete=models.CASCADE,
        verbose_name="Portfolio",
        help_text="The Portfolio the Asset belongs to.",
    )

    portfolio_created = models.ForeignKey(
        to="wbportfolio.Portfolio",
        null=True,
        blank=True,
        related_name="assets_created",
        on_delete=models.CASCADE,
        verbose_name="Portfolio Created",
        help_text="The Portfolio that created the Asset.",
    )

    ########################################################
    #                       CURRENCY                       #
    ########################################################

    currency = models.ForeignKey(
        to="currency.Currency",
        related_name="portfolio_currencies_asset",
        on_delete=models.PROTECT,
        verbose_name="Currency",
        help_text="The Currency of the Asset.",
    )

    currency_fx_rate_instrument_to_usd = models.ForeignKey(
        to="currency.CurrencyFXRates",
        related_name="instrument_assets",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name="Instrument Currency Rate",
        help_text="Rate to between instrument currency and USD",
    )
    currency_fx_rate_portfolio_to_usd = models.ForeignKey(
        to="currency.CurrencyFXRates",
        related_name="portfolio_assets",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name="Portfolio Currency Rate",
        help_text="Rate to between portfolio currency and USD",
    )
    ########################################################
    #                      Underlying                      #
    ########################################################

    underlying_instrument = models.ForeignKey(
        to="wbfdm.Instrument",
        related_name="instrument_assets",
        on_delete=models.PROTECT,
        verbose_name="Underlying Instrument",
        help_text="The instrument that is this asset.",
    )

    underlying_quote = models.ForeignKey(
        to="wbfdm.Instrument",
        related_name="assets",
        limit_choices_to=models.Q(children__isnull=True),
        on_delete=models.PROTECT,
        verbose_name="Underlying Quote",
        help_text="The quote that is this asset.",
    )

    underlying_quote_price = models.ForeignKey(
        to="wbfdm.InstrumentPrice",
        related_name="assets",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name="Underlying Instrument Price",
        help_text="The instrument price that is this asset.",
    )
    # objects = models.Manager()
    objects = DefaultAssetPositionManager()
    analytical_objects = AnalyticalAssetPositionManager()
    unannotated_objects = models.Manager()

    def pre_save(
        self, create_underlying_quote_price_if_missing: bool = False, infer_underlying_quote_price: bool = True
    ):
        if not self.asset_valuation_date:
            self.asset_valuation_date = self.date

        if (
            (not hasattr(self, "underlying_instrument") or not self.underlying_instrument)
            and hasattr(self, "underlying_quote")
            and self.underlying_quote
        ):
            self.underlying_instrument = (
                self.underlying_quote.parent if self.underlying_quote.parent else self.underlying_quote
            )
        elif (
            hasattr(self, "underlying_instrument")
            and self.underlying_instrument
            and (not hasattr(self, "underlying_quote") or not self.underlying_quote)
        ):
            try:
                self.underlying_quote = self.underlying_instrument.children.get(is_primary=True)
            except:
                self.underlying_quote = self.underlying_instrument

        if not getattr(self, "currency", None):
            self.currency = self.underlying_quote.currency
        if not self.underlying_quote_price and (infer_underlying_quote_price or not self.initial_price):
            try:
                # We get only the instrument price (and don't create it) because we don't want to create product instrument price on asset position propagation
                # Instead, we decided to opt for a post_save based system that will assign the missing position price when a price is created
                self.underlying_quote_price = InstrumentPrice.objects.get(
                    calculated=False, instrument=self.underlying_quote, date=self.asset_valuation_date
                )
            except InstrumentPrice.DoesNotExist:
                # if we create instrument price automatically, we need to ensure that the position is not estimated and not from a fake portfolio (e.g. JPM morgan root portfolio)
                if create_underlying_quote_price_if_missing and not self.is_estimated:
                    net_value = self.initial_price
                    # in case the position currency and the linked underlying_quote currency don't correspond, we convert the rate accordingly
                    if self.currency != self.underlying_quote.currency:
                        net_value *= self.currency.convert(self.asset_valuation_date, self.underlying_quote.currency)
                    self.underlying_quote_price = InstrumentPrice.objects.create(
                        calculated=False,
                        instrument=self.underlying_quote,
                        date=self.asset_valuation_date,
                        net_value=net_value,
                        import_source=self.import_source,  # we set the import source to know where this price is coming from
                    )
                    self.underlying_quote_price.fill_market_capitalization()
                    self.underlying_quote_price.save()
                else:  # sometime, the asset valuation date does not correspond to a valid market date. In that case, we get the latest valid instrument price for that product
                    self.underlying_quote_price = (
                        InstrumentPrice.objects.filter(
                            calculated=False,
                            instrument=self.underlying_quote,
                            date__lte=self.asset_valuation_date,
                        )
                        .order_by("date")
                        .last()
                    )

        if not self.currency_fx_rate_instrument_to_usd or (self.currency_fx_rate_instrument_to_usd.date != self.date):
            with suppress(CurrencyFXRates.DoesNotExist):
                self.currency_fx_rate_instrument_to_usd = CurrencyFXRates.objects.get(
                    date=self.date, currency=self.underlying_quote.currency
                )

        if not self.currency_fx_rate_portfolio_to_usd or (self.currency_fx_rate_portfolio_to_usd.date != self.date):
            with suppress(CurrencyFXRates.DoesNotExist):
                self.currency_fx_rate_portfolio_to_usd = CurrencyFXRates.objects.get(
                    date=self.date, currency=self.portfolio.currency
                )

        if not self.initial_price and self.underlying_quote_price:
            self.initial_price = self.underlying_quote_price.net_value
        if self.initial_currency_fx_rate is None:
            self.initial_currency_fx_rate = Decimal(1.0)
            if self.currency_fx_rate_portfolio_to_usd and self.currency_fx_rate_instrument_to_usd:
                try:
                    self.initial_currency_fx_rate = (
                        self.currency_fx_rate_portfolio_to_usd.value / self.currency_fx_rate_instrument_to_usd.value
                    )
                except InvalidOperation:
                    self.initial_currency_fx_rate = Decimal(0.0)

        # we set the initial shares from the previous position shares number if portfolio allows it
        if self.initial_shares is None and not self.portfolio.only_weighting:
            with suppress(AssetPosition.DoesNotExist):
                previous_pos = AssetPosition.objects.get(
                    date=(self.date - BDay(1)).date(),
                    underlying_quote=self.underlying_quote,
                    portfolio=self.portfolio,
                    portfolio_created=self.portfolio_created,
                )
                self.initial_shares = previous_pos.initial_shares
        if self.underlying_quote:
            self.exchange = self.underlying_quote.exchange

    def save(self, *args, create_underlying_quote_price_if_missing: bool = False, **kwargs):
        self.pre_save(create_underlying_quote_price_if_missing=create_underlying_quote_price_if_missing)
        super().save(*args, **kwargs)

    class Meta:
        verbose_name = "Asset Position"
        verbose_name_plural = "Asset Positions"
        indexes = [
            models.Index(fields=["date", "underlying_instrument", "portfolio"]),
            models.Index(fields=["date", "underlying_instrument"]),
        ]
        constraints = [
            models.CheckConstraint(
                check=~models.Q(date__week_day__in=[1, 7]),
                name="%(app_label)s_%(class)s_weekday_constraint",
            ),
            models.UniqueConstraint(
                fields=["portfolio", "date", "underlying_quote", "portfolio_created"],
                name="unique_asset_position",
                nulls_distinct=False,
            ),
        ]

    def __str__(self):
        return f"{self.initial_price} - {self.initial_shares} ({self.date}) ({str(self.underlying_quote)})"

    def set_weighting(self, new_weighting: Decimal):
        # Use this method to set the new weighting and ensure that the relative shares are updated accordingly
        self.weighting = new_weighting
        if self.initial_shares is not None:
            if self.weighting == 0 or self.initial_shares == 0:
                self.initial_shares = new_weighting * self.get_portfolio_total_asset_value()
            else:
                self.initial_shares = (new_weighting / self.weighting) * self.initial_shares

    def get_portfolio_total_asset_value(self) -> Decimal:
        return self.portfolio.get_total_asset_value(self.date)

    def _build_dto(self, **kwargs) -> PositionDTO:
        """
        Data Transfer Object
        Returns:
            DTO position object
        """
        parameters = dict(
            underlying_instrument=self.underlying_quote.id,
            weighting=self.weighting,
            shares=self._shares,
            date=self.date,
            asset_valuation_date=self.asset_valuation_date,
            instrument_type=self.underlying_quote.security_instrument_type.id,
            currency=self.underlying_quote.currency.id,
            country=self.underlying_quote.country.id if self.underlying_quote.country else None,
            is_cash=self.underlying_quote.is_cash or self.underlying_quote.is_cash_equivalent,
            primary_classification=(
                self.underlying_quote.primary_classification.id
                if self.underlying_quote.primary_classification
                else None
            ),
            favorite_classification=(
                self.underlying_quote.favorite_classification.id
                if self.underlying_quote.favorite_classification
                else None
            ),
            market_capitalization_usd=self._market_capitalization_usd,
            market_share=self._market_share,
            daily_liquidity=self._liquidity,
            volume_usd=self._volume_usd,
            price=self._price,
            currency_fx_rate=self._currency_fx_rate,
            portfolio_created=self.portfolio_created.id if self.portfolio_created else None,
        )
        parameters.update(kwargs)
        return PositionDTO(**parameters)

    @cached_property
    @admin.display(description="Adjusting Factor (adjustment)")
    def _adjusting_factor(self) -> Decimal:
        return (
            self.applied_adjustment.cumulative_factor * self.applied_adjustment.factor
            if self.applied_adjustment
            else Decimal(1.0)
        )

    @cached_property
    @admin.display(description="Price (Portfolio)")
    def _shares(self) -> Decimal:
        if self.initial_shares:
            return self.initial_shares / self._adjusting_factor
        return Decimal(0)

    @cached_property
    @admin.display(description="Market Capitalization")
    def _market_capitalization(self) -> float:
        return self.underlying_quote_price.market_capitalization_consolidated if self.underlying_quote_price else None

    @cached_property
    @admin.display(description="Volume 50d")
    def _volume_50d(self) -> float:
        return self.underlying_quote_price.volume_50d if self.underlying_quote_price else None

    @cached_property
    @admin.display(description="Price (Instrument)")
    def _price(self) -> Decimal:
        return self.underlying_quote_price.net_value if self.underlying_quote_price else self.initial_price

    @cached_property
    @admin.display(description="FX rate")
    def _currency_fx_rate(self) -> Decimal:
        return (
            self.currency_fx_rate_portfolio_to_usd.value / self.currency_fx_rate_instrument_to_usd.value
            if (self.currency_fx_rate_portfolio_to_usd and self.currency_fx_rate_instrument_to_usd)
            else self.initial_currency_fx_rate
        )

    @cached_property
    @admin.display(description="Price (Portfolio)")
    def _price_fx_portfolio(self) -> Decimal:
        return self._price * self._currency_fx_rate

    @cached_property
    @admin.display(description="Total Value (Instrument)")
    def _total_value(self) -> Decimal:
        if self._shares is not None:
            return self._price * self._shares
        return Decimal(0)

    @cached_property
    def fx_usd(self) -> Decimal:
        if self.currency_fx_rate_instrument_to_usd:
            return self.currency_fx_rate_instrument_to_usd.value
        return Decimal(1.0)

    @cached_property
    @admin.display(description="Total Value (USD)")
    def _total_value_fx_usd(self) -> Decimal:
        if self._shares is not None:
            return self._price * self._shares * self.fx_usd
        return Decimal(0)

    @cached_property
    @admin.display(description="Total Value (Portfolio)")
    def _total_value_fx_portfolio(self) -> Decimal:
        if self._shares is not None:
            return self._price * self._shares * self._currency_fx_rate
        return Decimal(0)

    @cached_property
    @admin.display(description="Market Share")
    def _market_share(self) -> Decimal:
        if self._total_value is not None and self._market_capitalization:
            return self._total_value / Decimal(self._market_capitalization)
        return Decimal(0)

    @cached_property
    @admin.display(description="Liquidity")
    def _liquidity(self) -> float:
        if self._total_value is not None and self._volume_50d:
            return float(self._shares) / self._volume_50d / 0.33
        return 0.0

    @cached_property
    @admin.display(description="Market Capitalization (USD)")
    def _market_capitalization_usd(self) -> float:
        if self._market_capitalization is not None:
            return self._market_capitalization * float(self.fx_usd)
        return 0.0

    @cached_property
    @admin.display(description="Volume (USD)")
    def _volume_usd(self) -> float:
        if self._price_fx_portfolio is not None and self._volume_50d is not None:
            return float(self._price_fx_portfolio) * float(self.fx_usd) * self._volume_50d
        return 0.0

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbportfolio:assetposition"

    @classmethod
    def currency_group_by(cls, qs, field_name: str | None = "key"):
        return (
            qs.filter(currency__isnull=False)
            .values("currency")
            .annotate(
                groupby_id=models.F("currency__id"),
                aggregated_title=models.F(f"currency__{field_name}"),
            )
        )

    @classmethod
    def country_group_by(cls, qs, field_name: str | None = "name"):
        return (
            qs.filter(underlying_instrument__country__isnull=False)
            .values("underlying_instrument__country")
            .annotate(
                groupby_id=models.F("underlying_instrument__country__id"),
                aggregated_title=models.F(f"underlying_instrument__country__{field_name}"),
            )
        )

    @classmethod
    def exchange_group_by(cls, qs, field_name: str | None = "name"):
        return qs.values("exchange").annotate(
            groupby_id=models.F("exchange"),
            aggregated_title=models.F(f"exchange__{field_name}"),
        )

    @classmethod
    def cash_group_by(cls, qs, **kwargs):
        return (
            qs.annotate(
                underlying_security_instrument_type_name_repr=Case(  # Annotate the parent security if exists
                    When(
                        underlying_instrument__isnull=False,
                        then=F("underlying_instrument__instrument_type__name_repr"),
                    ),
                    default=F("underlying_instrument__instrument_type__name_repr"),
                ),
                is_cash=Case(
                    When(
                        Q(underlying_instrument__is_cash=True) | Q(underlying_instrument__is_cash_equivalent=True),
                        then=Value("Cash"),
                    ),
                    default=F("underlying_security_instrument_type_name_repr"),
                    output_field=CharField(),
                ),
            )
            .values("is_cash")
            .annotate(groupby_id=models.F("is_cash"), aggregated_title=models.F("is_cash"))
        )

    @classmethod
    def equity_group_by(cls, qs, field_name: str | None = "name"):
        return (
            qs.filter(underlying_instrument__isnull=False)
            .values("underlying_instrument")
            .annotate(
                groupby_id=models.F("underlying_instrument__id"),
                aggregated_title=models.F(f"underlying_instrument__{field_name}"),
            )
        )

    @classmethod
    def marketcap_group_by(cls, qs, **kwargs):
        qs = qs.filter(market_capitalization_usd__isnull=False).annotate(
            mktcap_allocation=Case(
                When(market_capitalization_usd__isnull=True, then=Value("None")),
                When(market_capitalization_usd__gt=MARKETCAP_XL, then=Value("> 300B")),
                When(
                    market_capitalization_usd__gt=MARKETCAP_L,
                    then=Value("50B to 300B"),
                ),
                When(
                    market_capitalization_usd__gt=MARKETCAP_M,
                    then=Value("10B to 50B"),
                ),
                When(
                    market_capitalization_usd__gt=MARKETCAP_S,
                    then=Value("2B to 10B"),
                ),
                default=Value("< 2B"),
                output_field=CharField(),
            )
        )
        return qs.values("mktcap_allocation").annotate(
            groupby_id=F("mktcap_allocation"), aggregated_title=F("mktcap_allocation")
        )

    @classmethod
    def liquidity_group_by(cls, qs, **kwargs):
        qs = qs.annotate(
            liquidity_allocation=Case(
                When(liquidity__isnull=True, then=Value("None")),
                When(
                    liquidity__gt=LIQUIDITY_LARGE,
                    then=Value(f"More than {int(LIQUIDITY_LARGE)} days to liquidate"),
                ),
                When(
                    liquidity__lt=LIQUIDITY_SMALL,
                    then=Value(f"Less than {int(LIQUIDITY_SMALL)} days to liquidate"),
                ),
                default=Value(f"{int(LIQUIDITY_SMALL)} to {int(LIQUIDITY_LARGE)} days to liquidate"),
                output_field=CharField(),
            ),
        )
        return qs.values("liquidity_allocation").annotate(
            groupby_id=F("liquidity_allocation"), aggregated_title=F("liquidity_allocation")
        )

    @classmethod
    def group_by_primary(cls, qs: models.QuerySet, height: int = 0, **kwargs):
        qs = (
            qs.annotate_classification_for_group(
                ClassificationGroup.objects.get(is_primary=True), classification_height=height, unique=True
            )
            .annotate(
                classification_id=F("classifications"),
                classification_title=Subquery(
                    Classification.objects.filter(id=OuterRef("classifications")).values("name")[:1]
                ),
            )
            .filter(classification_id__isnull=False)
        )
        return qs.values("classification_id").annotate(
            groupby_id=F("classification_id"), aggregated_title=F("classification_title")
        )

    @classmethod
    def group_by_preferred_classification(cls, qs: models.QuerySet, height: int, **kwargs):
        qs = qs.annotate_preferred_classification_for_group(
            ClassificationGroup.objects.get(is_primary=True), classification_height=height
        )
        return qs.values("classification_id").annotate(
            id=F("classification_id"), aggregated_title=F("classification_title")
        )

    @classmethod
    def get_shown_positions(cls, person):
        from wbportfolio.models.portfolio import Portfolio

        today = date.today()
        if person.user_account.is_superuser:
            return AssetPosition.objects.all()
        else:
            portfolios = Portfolio.objects.annotate(
                nb_roles=Coalesce(
                    Subquery(
                        PortfolioRole.objects.filter(
                            (
                                (
                                    Q(person=person)
                                    & Q(
                                        role_type__in=[
                                            PortfolioRole.RoleType.MANAGER,
                                            PortfolioRole.RoleType.RISK_MANAGER,
                                        ]
                                    )
                                )
                                | (
                                    Q(person=person)
                                    & Q(
                                        role_type__in=[
                                            PortfolioRole.RoleType.PORTFOLIO_MANAGER,
                                            PortfolioRole.RoleType.ANALYST,
                                        ]
                                    )
                                    & (Q(instrument=OuterRef("instruments")) | Q(instrument__isnull=True))
                                )
                            )
                            & (Q(start__isnull=True) | Q(start__lte=today))
                            & (Q(end__isnull=True) | Q(end__gte=today))
                        )
                        .annotate(c=models.Count("*"))
                        .values("c")[:1]
                    ),
                    0,
                )
            ).filter(nb_roles__gt=0)
            return AssetPosition.objects.filter(portfolio__id__in=portfolios.values_list("id", flat=True))

    @classmethod
    def get_invested_instruments(cls, only_on_date: date, portfolio=None):
        product_portfolios = InstrumentPortfolioThroughModel.objects.filter(instrument__instrument_type__key="product")
        if portfolio:
            product_portfolios = product_portfolios.filter(portfolio=portfolio)
        asset_positions = AssetPosition.objects.filter(
            portfolio__in=product_portfolios.values("portfolio"), date=only_on_date
        )
        return (
            Instrument.annotated_objects.filter(is_investable_universe=True)
            .annotate(has_position=Exists(asset_positions.filter(underlying_quote=OuterRef("id"))))
            .filter(has_position=True)
        )


@receiver(post_save, sender="wbfdm.InstrumentPrice")
def post_instrument_price_creation(sender, instance, created, raw, **kwargs):
    if not raw and created and not instance.calculated:
        AssetPosition.objects.filter(
            Q(asset_valuation_date=instance.date)
            & Q(underlying_quote=instance.instrument)
            & (Q(underlying_quote_price__isnull=True) | ~Q(asset_valuation_date=F("underlying_quote_price__date")))
        ).update(underlying_quote_price=instance)


@receiver(pre_merge, sender="wbfdm.Instrument")
def pre_merge_instrument(sender: models.Model, merged_object: Instrument, main_object: Instrument, **kwargs):
    """
    Simply reassign the instrument price linked to the merged instrument to the main instrument if they don't already exist. Otherwise, delete them
    """
    merged_object.assets.annotate(
        new_price=InstrumentPrice.objects.filter(
            instrument=main_object, date=OuterRef("date"), calculated=False
        ).values("id")[:1]
    ).update(
        underlying_quote=main_object,
        underlying_instrument=main_object.parent if main_object.parent else main_object,
        underlying_quote_price=F("new_price"),
    )


@receiver(add_instrument_to_investable_universe, sender="wbfdm.Instrument")
def add_instrument_to_investable_universe(sender: models.Model, **kwargs) -> list[int]:
    """
    register all instrument linked to assets as within the investible universe
    """
    return list(
        (
            Instrument.objects.annotate(
                assets_exists=Exists(
                    AssetPosition.objects.filter(portfolio__is_tracked=True, underlying_quote=OuterRef("pk"))
                )
            ).filter(Q(assets_exists=True) | Q(portfolios__isnull=False))
        )
        .distinct()
        .values_list("id", flat=True)
    )
